program pascalmdis ;

{$APPTYPE CONSOLE}

uses
  SysUtils;

(* disassembler for M-pascal object files V1.4 *)

(* Written by   : H.J.C. Otten
   Last update  : 11 aug 2020

   Code cleaned up, no more warnings
   M-code not essentially changed from V1.0
   GFA and SFA and MVB instruction already in V1.0
   As Standard procedures RSET 'REWR ' 'STATE' CLOSE' ;
 *)

CONST

  defaultkim = $2FA3 ; (* according to V1.0 KIM-1 Pascal-M interpreter *)

  alphalength = 80 ;
  maxstore = 32767 ;
  maxproc  = 99   ;
  maxpage  =  56   ;
  {
  maxlabel = 3000  ;
  }
  maxstandardprocedures = 16 ;
  strglgth = 8 ; (* length of ident, procedure/program name *)

TYPE
{
  phasetype = (phase1, phase2) ;
}
  filestatus = ( correct, endoffile, overflowbuffer, checkerror ) ;

  alpha = packed array [ 1 .. 80] of char ;

  procinfo = record
               procaddress : integer ;
               procname    : packed array[1..strglgth] of char ;
              END ;


  instructname = packed array[1 .. 5] of char ;

  instruct = record
               name : instructname ;
               arg  : integer
             END ;


VAR

  (* objectfile Variables *)
  linelength, recnr    : integer ;
  checksum, rchecksum  : integer ;
  fstatus               : filestatus ;
  line  : alpha ;

  (* procedure information buffer *)
  procbuffer : array [ 0 .. maxproc ] of procinfo ;
  procf  : integer ; (* total nr of procedures loaded   *)
{
  labels  : array [1..maxlabel] of integer ;

  (* addresses of label, value > 0 means label used *)
  labelcount : integer ; (* last label added to label array *)
}

  (* p-machine memory *)
  store : array [0 .. maxstore] of byte ;

  (* registers of p-machine *)
  pc     : integer ; (*  nr of code bytes laoded in store *)
  ldaddr : integer ; (* offset *)

  (* disassembler counter *)
  dpc : integer ;

  (* files *)
  objectfile, asmfile, listfile : text ;

  filename,
  listfilename,
  asmfilename,
  filenamepart: string ;


  (* instructions with level in instruction byte, 0 - $8F  *)
  grinstruction : array [0 .. 8] of instruct ;

  (* instructions above $8F , whole byte unique *)
  instruction : array [$90 .. $FF] of instruct ;

  (* Standard procedures *)
  standardproc : array [0..maxstandardprocedures] of string;
  (* Standard procedures comment *)
  standardcproc : array [0..maxstandardprocedures] of string;

  (* listing file counters *)
  listlines, listpages : integer ;

  objectlinepre,
  objectline   : string ;
  assemblerline : string ;

  PROCEDURE EndLinelist  ;
  (**********************)

    (* Ends line on listfile takes care of pages, headings etc *)


    BEGIN

      listlines := listlines + 1 ;
      (* new page needed ? *)
      IF listlines > maxpage
        THEN
          BEGIN
            (* write formfeed and header *)
            writeln ( listfile ) ;
            write ( listfile, 'Pascal-M V1.0 Disassembler   ') ;
            write ( listfile, filename) ;
            writeln ( listfile, '               Page ', listpages:4 ) ;
            writeln ( listfile ) ;
            writeln ( listfile ) ;
            listpages := listpages + 1 ;
            listlines := 3
          END ;

    END ; (* EndLineList *)


  PROCEDURE FillInstructions  ;
  (* fills instruction table with names
       of number of arguments              *)

  VAR

    i : integer ;

    PROCEDURE FillGrinstruction(index: integer ; name: instructname ; arg: integer ) ;

      BEGIN

        grinstruction[index].name := name ;
        grinstruction[index].arg := arg ;

      END ; (* Fillgrinstructions *)


    PROCEDURE fillinstruct(index: integer ; name: instructname ; arg: integer ) ;

      BEGIN

        instruction[index].name := name ;
        instruction[index].arg := arg ;

      END ; (* FillInstruct *)

    BEGIN (* FillInstructions *)

        (* instructions 0 -$8F = 143, contain level in instruction *)
	Fillgrinstruction(0,'LDCIS',0)        ;
	Fillgrinstruction(1,'LDAS ',1)        ;
	Fillgrinstruction(2,'LDA  ',2)        ;
	Fillgrinstruction(3,'MST0 ',0)        ;
	Fillgrinstruction(4,'MSTN ',1)        ;
	Fillgrinstruction(5,'LOD1 ',1)        ;
	Fillgrinstruction(6,'LOD2 ',1)        ;
	Fillgrinstruction(7,'STR1 ',1)        ;
	Fillgrinstruction(8,'STR2 ',1)        ;

  fillinstruct($90,'LEQ2 ',0)       ;
	fillinstruct($91,'FOR  ',0)       ;
	fillinstruct($92,'LEQM ',2)       ;
	fillinstruct($93,'LES2 ',0)       ;
	fillinstruct($94,'LEQ8 ',0)       ;
	fillinstruct($95,'LESM ',2)       ;
	fillinstruct($96,'EQU2 ',0)       ;
	fillinstruct($97,'GEQ8 ',0)       ;
	fillinstruct($98,'EQUM ',2)       ;
	fillinstruct($99,'EQU8 ',0)       ;
	fillinstruct($9A,'IND1 ',0)       ;
	fillinstruct($9B,'IND2 ',0)       ;
	fillinstruct($9C,'IND8 ',0)       ;
	fillinstruct($9D,'STO1 ',0)       ;
	fillinstruct($9E,'ST02 ',0)       ;
	fillinstruct($9F,'STO8 ',0)       ;
	fillinstruct($A0,'LDC  ',2)       ;
	fillinstruct($A1,'RETP ',0)       ;
	fillinstruct($A2,'ADI  ',0)       ;
	fillinstruct($A3,'AND  ',0)       ;
	fillinstruct($A4,'DIF  ',0)       ;
	fillinstruct($A5,'DVI  ',0)       ;
	fillinstruct($A6,'INN  ',0)       ;
	fillinstruct($A7,'INT  ',0)       ;
	fillinstruct($A8,'IOR  ',0)       ;
	fillinstruct($A9,'MOD  ',0)       ;
	fillinstruct($AA,'MPI  ',0)       ;
	fillinstruct($AB,'NGI  ',0)       ;
	fillinstruct($AC,'NOT  ',0)       ;
	fillinstruct($AD,'SBI  ',0)       ;
	fillinstruct($AE,'SGS  ',0)       ;
	fillinstruct($AF,'UNI  ',0)       ;
	fillinstruct($B0,'LNC  ',2)       ;
	fillinstruct($B1,'FJP  ',2)       ;
	fillinstruct($B2,'UJP  ',2)       ;
	fillinstruct($B3,'DEC  ',2)       ;
	fillinstruct($B4,'INC  ',2)       ;
	fillinstruct($B5,'ENT  ',2)       ;
	fillinstruct($B6,'CAS  ',0)       ;
	fillinstruct($B7,'MOV  ',2)       ;
	fillinstruct($B8,'DEC1 ',0)       ;
	fillinstruct($B9,'INC1 ',0)       ;
	fillinstruct($BA,'LDCS ',8)       ;
	fillinstruct($BB,'CAP  ',2)       ;
	fillinstruct($BC,'LCA  ',5)       ; (* 5 is flag for writing arguments *)
	fillinstruct($BD,'CSP  ',1)       ;
	fillinstruct($BE,'CUP1 ',1)       ;
	fillinstruct($BF,'CUP2 ',1)       ;
	fillinstruct($C0,'FIX21',0)       ;
	fillinstruct($C1,'LNS  ',0)       ;
	fillinstruct($C2,'SFA  ',0)       ;
	fillinstruct($C3,'GFA  ',0)       ;
	fillinstruct($C4,'MVB  ',4)       ;

	FOR i := $C5 to $FF DO
	  BEGIN
	    instruction [i].name := 'NOP  '  ;
            instruction [i].arg  := 0        ;
	  END ;

	standardproc[ 0] := 'WRI  ' ;
	standardproc[ 1] := 'WRC  ' ;
	standardproc[ 2] := 'WRS  ' ;
	standardproc[ 3] := 'RDI  ' ;
	standardproc[ 4] := 'RDL  ' ;
	standardproc[ 5] := 'RDC  ' ;
	standardproc[ 6] := 'WLN  ' ;
	standardproc[ 7] := 'NEW  ' ;
	standardproc[ 8] := 'EOF  ' ;
	standardproc[ 9] := 'RST  ' ;
	standardproc[10] := 'ELN  ' ;
	standardproc[11] := 'STP  ' ;
	standardproc[12] := 'ODD  ' ;
	standardproc[13] := 'RSET ' ;
        standardproc[14] := 'REWR ' ;
        standardproc[15] := 'STATE' ;
        standardproc[16] := 'CLOSE' ;


        standardcproc[ 0] := 'write integer'  ;
        standardcproc[ 1] := 'write character' ;
        standardcproc[ 2] := 'write string'  ;
        standardcproc[ 3] := 'read integer'  ;
        standardcproc[ 4] := 'read to end of line'  ;
        standardcproc[ 5] := 'read character' ;
        standardcproc[ 6] := 'write end of line'  ;
        standardcproc[ 7] := 'new' ;
        standardcproc[ 8] := 'end of file test'  ;
        standardcproc[ 9] := 'restore heap' ;
        standardcproc[10] := 'end of line test';
        standardcproc[11] := 'stop execution' ;
        standardcproc[12] := 'test integer for odd' ;
        standardcproc[13] := 'reset file'   ;
        standardcproc[14] := 'rewrite file' ;
        standardcproc[15] := 'status file' ;
        standardcproc[16] := 'close file'  ;

    END ; (* FillInstructions *)

  PROCEDURE Loader ;
  (* loads p-code output from compiler *)
  (* records output as followed by checksum
     P1 = code-bytes
     P2 = address-reference four bytes
     P4 = procedure level + procedure name
     P9 = END of code ( no checksum )       *)

VAR

  count, i : integer ;


  PROCEDURE GetLine ( VAR linelength : integer ;
  (***************)   VAR fstatus : filestatus ) ;

    BEGIN

     IF eof(objectfile )
       THEN
         fstatus := endoffile
       ELSE
         BEGIN
           recnr := recnr + 1 ;
           linelength := 1 ;
           read (objectfile, line [ linelength ] ) ;
           while not eoln(objectfile ) and ( linelength < alphalength ) do
             BEGIN
               linelength := linelength + 1 ;
               read ( objectfile, line [ linelength ] )  ;
             END ;
           IF not eoln (objectfile )
             THEN
               fstatus := overflowbuffer ;
           readln (objectfile ) ;
         END ;

    END ; (* GetLine *)

  FUNCTION PackHex(startposition : integer ) : integer ;

  VAR

    hexnumber, chvalue, i, j  : integer ;

  BEGIN

   hexnumber := 0 ;
   for i := 0 to 1 do
     BEGIN
       chvalue := ord ( line[startposition + i ] ) ;
       IF ( chvalue >= ord ('0') ) and
          ( chvalue <= ord ('9') )
         THEN
           j := chvalue - ord ('0')
         ELSE
           j := chvalue - ord ('A') +  10 ;
       hexnumber := hexnumber * 16 + j
      END ;
    PackHex := hexnumber ;

  END ; (* PackHex *)

  PROCEDURE Loadp1 ;
  (* loads P1 record into procedure buffer *)

  VAR

    i,nrloadbytes : integer ;

  BEGIN

    nrloadbytes := PackHex ( 3 ) ;
    checksum := nrloadbytes ;
    for i := 1 to nrloadbytes - 1 do
      BEGIN
        store[pc] := PackHex (  2*i - 1 + 4  ) ;
        checksum := checksum + store[pc] ;
        pc := pc + 1 ;
      END ;
    rchecksum := PackHex ( nrloadbytes * 2 - 1 + 4 ) ;
    IF ( ( 16383 - checksum ) mod 256 ) <> rchecksum
      THEN
        BEGIN
          writeln ('Checksum error in P1 record ') ;
          fstatus := checkerror
        END ;

  END ; (* Loadp1 *)

  PROCEDURE Loadp2 ;
  (* loads P2 record with address into store buffer *)

  VAR

     loc, val, i, locl, loch, vall, valh  : integer ;

   BEGIN

     loch := PackHex(3) ;
     locl := PackHex(5) ;
     loc :=  (256* loch +  locl) + ldaddr   ;
     checksum := loch + locl ;
     valh := PackHex(7) ;
     vall := Packhex(9) ;
     val := (256 * valh + vall)  ;
     checksum := checksum +  valh + vall ;
     store [ loc   ]  :=  val div 256 ;
     store [ loc + 1] := val mod 256  ;
     rchecksum := PackHex ( 11 ) ;
     i := ( 16383 - checksum ) mod 256 ;
     if i <> rchecksum
       then
         BEGIN
           writeln ('Checksum error in P2 record ') ;
           fstatus := checkerror
         END ;

  END ; (* Loadp2 *)

  PROCEDURE Loadp4 ;
  (* loads P4 record into PROCEDURE buffer   *)
  (* line [3] .. line[4] is PROCEDURE number *)
  (* line [5] .. line [5 + strglgth] is PROCEDURE name *)

  VAR

    nrproc, i : integer ;

  BEGIN

    procf := procf + 1 ;
    nrproc := PackHex(3) ;
    checksum := nrproc ;
    for i := 1 to strglgth do
     BEGIN
      procbuffer[nrproc].procname[i] := line[4 + i] ;
      checksum := checksum + ord ( line[i+4] ) ;
     END ;
    procbuffer[nrproc].procaddress := pc  ;
    rchecksum := PackHex ( 13 ) ;
    IF ( ( 16383 - checksum ) mod 256 ) <> rchecksum
      THEN
        BEGIN
          writeln ('Checksum error in P4 record ') ;
          fstatus := checkerror
        END ;

  END ; (* Loadp4 *)

  BEGIN (* loader *)

    for count := 0 to maxproc do
      for i := 1 to strglgth do
        procbuffer[count].procname[i] := ' ' ;

    pc := ldaddr ;       (* Store starts at load address  *)

    fstatus := correct ;
    recnr := 0 ;
    procf := 0 ;
    repeat
      getline ( linelength, fstatus ) ;
      checksum := 0 ;
      IF fstatus <> endoffile
        THEN
          IF ( line[1] = 'P' ) or ( line[1] = 'p' )
            THEN
              case line[2] of
               '1' : Loadp1 ;
               '2' : Loadp2 ;
               '4' : Loadp4 ;
               '9' : fstatus := endoffile
              END ;

    until (fstatus = endoffile ) or ( fstatus = checkerror ) ;

    writeln ;
    IF fstatus = checkerror
      THEN
        begin
          writeln ('Loading ended due to checkerror in record : ', recnr ) ;
          halt(1) ;
        end ;

  END ; (* Loader *)

function WriteHex(hexbyte : integer): string ;

  (* write byte as two hex numbers into hexstring *)

  VAR

    hexstring: string ;

  function nibbletohex(nibble: integer): char ;

    BEGIN

      IF nibble  < 10
        THEN
          nibbletohex := chr(ord ('0') + nibble)
        ELSE
          nibbletohex := chr (ord ('A') + nibble - 10 )  ;

     end ; (* nibbletohex *)

  BEGIN

    if (hexbyte >= 0) and (hexbyte <= 256)
       then
         BEGIN
           hexstring := nibbletohex(hexbyte div 16) ;
           hexstring := hexstring + nibbletohex(hexbyte mod 16) ;
         end
       else
         hexstring := '  ' ;
    WriteHex := hexstring ;

  END ; (* WriteHex *)


PROCEDURE WriteLevelInstruction ( opcode  : integer ; var dpc : integer ) ;
    (* Prints instruction mnemonic and if necessary arguments as numbers *)

  VAR

    i, level : integer ;

  BEGIN

     level := opcode mod 256 ;
     opcode := opcode div 256 ;
     assemblerline := assemblerline + grinstruction[opcode].name +
      ' ' + WriteHex(level) ;
     FOR i := 1 to grinstruction[opcode].arg DO
       BEGIN
         objectline :=  objectline + ' ' + WriteHex(store[dpc + i]) ;
         assemblerline := assemblerline +  ' ' + WriteHex(store[dpc + i]) ;
       end ;
     dpc := dpc + 1 + grinstruction[opcode].arg ;

  END ; (* WriteLevelInstruction *)

PROCEDURE FindProc ( VAR procnr : integer ; pc : integer  ) ;
  (* Searches for proc/function in procbuffer with pc-address *)

  VAR
      i : integer ;

  BEGIN

     procnr := -1 ; (* non-existing proc nr *)
     i := 0 ;

     REPEAT
       IF procbuffer[i].procaddress = pc
        THEN
          BEGIN
            procnr := i ;
            i := procf + 1
          END
        ELSE
          i := i + 1
     UNTIL ( i > procf ) ;

  END ; (* FindProc *)

{
PROCEDURE FindLabel (address: integer ;  VAR labelnr : integer)   ;

  (* Searches if address is a label, index in label table is label number *)
  (* returns label number or -1 if label not found *)

  VAR
      i : integer ;

  BEGIN

     labelnr := -1 ; (* no label found *)
     i := 1 ;

     WHILE (i <= maxlabel) or (labelnr = -1) DO
       BEGIN
         IF labels[i] = address
           THEN
             labelnr := i ;
         i := i + 1
       END ;

  END ;

}
PROCEDURE WriteInstruction ( opcode  : integer ;
(************************)   var dpc : integer ) ;

    (* Prints instruction mnemonic and if necessary
       arguments as numbers or string *)

    VAR

      i,
      nrargs      : integer ;


    PROCEDURE WriteENT(dpc : integer) ;

      VAR

        i, procnr : integer ;

      BEGIN

        procnr := -1 ;
        FindProc (procnr, dpc ) ;
        IF procnr > -1
          THEN
            BEGIN
              write ( listfile, 'Procedure/Function/Program ' ) ;
              for i := 1 to strglgth do
                write ( listfile, procbuffer[procnr].procname[i] ) ;
              writeln (listfile )
            END
          ELSE
            writeln ( listfile, ' ':16, 'Unknown entry point') ;
        writeln(listfile) ;
        EndLineList ;

     END ; (* WriteENT *)

  PROCEDURE PRCASE ;
    (* dump CASE structure in code *)

    VAR
      i,j, nrlabels, startvalue : integer ;

    BEGIN
      writeln ( listfile ) ;
      EndLineList ;
      WriteHex ( ( dpc + 1 ) div 256 ) ;
      WriteHex ( ( dpc + 1 ) mod 256 ) ;
      write ( listfile, ' : ' ) ;
      WriteHex ( store [ dpc + 1 ] ) ;
      write ( listfile, ' ' ) ;
      WriteHex ( store [ dpc + 2 ] ) ;
      write ( listfile, '    start value case label = ' ) ;
      write ( listfile, (store [dpc + 1] * 256 + store [dpc + 2]):3 ) ;
      writeln ( listfile ) ;
      EndLineList ;
      WriteHex ( ( dpc + 3 ) div 256 ) ;
      WriteHex ( ( dpc + 3 ) mod 256 ) ;
      write ( listfile, ' : ' ) ;
      WriteHex ( store [ dpc + 3 ] ) ;
      write ( listfile, ' ' ) ;
      WriteHex ( store [ dpc + 4 ] ) ;
      write ( listfile, '      end value case label = ' ) ;
      write ( listfile, (store [dpc + 3] * 256 + store [dpc + 4] ):3 ) ;
      writeln ( listfile ) ;
      EndLineList ;
      WriteHex ( ( dpc + 5 ) div 256 ) ;
      WriteHex ( ( dpc + 5 ) mod 256 ) ;
      write ( listfile, ' : ' ) ;
      WriteHex ( store [ dpc + 5 ] ) ;
      write ( listfile, ' ' ) ;
      WriteHex ( store [ dpc + 6 ] ) ;
      j := store [dpc + 5] * 256 + store [dpc + 6] ;
      IF j = 0
        THEN
          write ( listfile, '    no otherwise part in case' )
        ELSE
          BEGIN
            write ( listfile, '    address of otherwise part = ' ) ;
            j := j + ldaddr ;
            WriteHex ( j div 256 ) ;
            WriteHex ( j mod 256 )
          END ;

      startvalue := store [dpc + 1] * 256 + store [dpc + 2] ;
      nrlabels := store [dpc + 3] * 256 + store [dpc + 4] -
                  startvalue ;

      FOR i := 0 TO  nrlabels DO
        BEGIN
          writeln ( listfile ) ;
          j := dpc + 2*i + 7 ;
          EndLineList ;
          WriteHex ( j div 256 ) ;
          WriteHex ( j mod 256 ) ;
          write ( listfile,  ' : ' ) ;
          WriteHex ( store [ j ] ) ;
          write ( listfile, ' ' ) ;
          WriteHex ( store [ j + 1 ] ) ;
          write ( listfile, '    case label value ' ) ;
          write ( listfile, (i + startvalue):3 ) ;
          j := store[ j ] * 256 + store [j + 1] ;
          IF j = 0
            THEN
              write ( listfile, ' not used')
            ELSE
              BEGIN
                write ( listfile, ' at address = ' ) ;
                j := j + ldaddr ;
                WriteHex ( j div 256 ) ;
                WriteHex ( j mod 256 )
              END ;
        END ;
      nrargs := 7 + (nrlabels + 1) * 2 ;

    END ; (* PRCASE *)


  PROCEDURE PrStringLCA ;
    (* Prints contents of string following opcode LCA
       in multiple lines if necessary                  *)

     VAR
       i, kar: integer ;

     BEGIN
       (* show string as hex bytes in objectfile *)
       for i := 1 to store [ dpc + 1 ] + 1 do
           objectline := objectline + ' ' + WriteHex(store[dpc + i]) ;
       (* show string as pascal string *)
       assemblerline := assemblerline + '''' ;
       FOR i:= 2 to  store [ dpc + 1 ] + 1 do
         BEGIN
           kar := store[dpc + i] ;
           IF kar < 31
             THEN
               kar := ord('^') ;
            assemblerline := assemblerline + chr(kar)
          END ;
       assemblerline := assemblerline + '''' ;
       nrargs := 1 + store [ dpc + 1 ] ;

    END ; (* PrStringLCA *)

    BEGIN (* WriteInstruction *)

      nrargs := instruction [opcode].arg ;
      assemblerline := assemblerline + instruction[opcode].name ;
      IF nrargs = 5
        THEN
          PrStringLCA
        ELSE
          IF opcode <> $BD
            THEN

              BEGIN
                (* dump arguments *)
                FOR i := 1 TO nrargs DO
                  BEGIN
                   assemblerline := assemblerline  + WriteHex(store[dpc + i]) + ' ';
                   objectline := objectline + WriteHex(store[dpc + i]) + ' ' ;
                  END ;
              END ;


      (* UJP and FJP *)

      IF (opcode = $B1) or (opcode = $B2)  (*  FJP or UJP *)
        THEN
          { should be label, collected in phase 1 }
          BEGIN
            i := store[dpc + 1] * 256 + store[dpc + 2] + ldaddr ;
            objectline := objectline + ' (* ' +
                          WriteHex(i div 256) + WriteHex(i mod 256)+  ' *)'
          END

      {
      ELSE IF (opcode = $91) (* FOR *)
        THEN

          BEGIN
            write (listfile, ' loop back to body ' ) ;
            i := store[dpc + 1] * 256 + store[dpc + 2] + ldaddr ;
            WriteHex ( i div 256 ) ;
            WriteHex ( i mod 256 )
          END
      }
      ELSE IF (opcode = $BD) (* CSP *)
        (* should be symbol output for assembler *)
        THEN
          BEGIN
            assemblerline := assemblerline + standardproc[store[dpc+1]] ;
            objectline := objectline + WriteHex(store[dpc+1])                         ;
          END
      {
      ELSE IF ( opcode = 47) OR (opcode = 48)
        THEN
          BEGIN
             procnr := store[ dpc + 1 ] ;
             write ( listfile, ' ', 'Procedure/Function/Program ' ) ;
             for i := 1 to 8 do
               write ( listfile, procbuffer[procnr].procname[i] ) ;
          END } ;

      dpc := dpc + 1 + nrargs ;

    END; (* WriteInstruction *)


  PROCEDURE DisAssemble {(phase: phasetype)} ;
  (* two phases
     - first one find labels, no output
     - second one, output with labels

     disassemble store upto pc last loaded

     asm file format

     list file format (normal instruction with two arguments max)

     AAAA OP (optional args) XX YY label opcode-menmonic values

     More then two args: line break after binary args

     CASE: special format

     UJP, FJP : label as value

     LCA string: string  *)

  VAR

    opcode  : integer ;

  BEGIN

    dpc := ldaddr ;

    EndLineList ;
    (* header for assembler file *)
    writeln(asmfile, '      .program ', procbuffer[0].procname) ;
    writeln(listfile, '     .program ', procbuffer[0].procname) ;

    REPEAT

      opcode := store[dpc] ;
      (* write address and opcode in hex *)
      objectlinepre := WriteHex(dpc div 256) + WriteHex(dpc mod 256) + ': '
                      + WriteHex(opcode) + '  ' ;
      objectline := '' ;
      (* label, now empty *)
      assemblerline := '      ' ;

      (* write arguments of opcode, label, values etc,
      increment dpc to next opcode *)
      IF (opcode < $90)
        THEN
          writelevelinstruction(opcode, dpc)
        ELSE
          WriteInstruction(opcode, dpc) ;
      writeln(listfile, objectlinepre , objectline) ;
      writeln(listfile, assemblerline) ;
      writeln(asmfile, assemblerline) ;
      EndLineList ;
    UNTIL dpc >= pc ;
    writeln(asmfile,'end.') ;

  END ; (* DisAssemble *)

PROCEDURE DumpProcedureBuffer ;

 var

   count : integer ;

begin

  listlines := maxpage + 1 ;
  EndLineList ;
  writeln(listfile, 'Procedure buffer') ;
  EndLIneList ;
  writeln(listfile, 'Number Address Name') ;
  EndLIneList ;
  writeln(listfile, '-------------------') ;
  EndLineList ;

  for count := 0 to procf do
    begin

      write(listfile, count:2, ' ', hexstr(procbuffer[count].procaddress, 4)) ;
      if procbuffer[count].procname[1] = ' '
        then
          writeln(listfile, ' external routine')
        else
          writeln(listfile,' ', procbuffer[count].procname) ;
       EndlineList ;
    end;

end;


PROCEDURE OpenFiles ;

  VAR

    i : integer ;

BEGIN (* Openfiles *)

  if paramcount = 0
    then
      begin
        write('Pascal-M object file to load> ') ;
        readln(filename)
      end
    else
      filename := paramstr(1) ;
  ldaddr := defaultkim ;
  if paramcount = 2
    then
      begin
        Try
          ldaddr := StrToInt(paramstr(2));
        except
          On E : EConvertError do
              ldaddr := defaultkim ;
         end;
       end;
  if FileExists(filename)
    then
      begin
        assign ( objectfile, filename);
        reset(objectfile) ;
      end
     else
       begin
         writeln('No such file') ;
         halt(1) ;
       end;

  { extract filename part }
  filenamepart := '' ;
  i := 1 ;
  while filename[i] <> '.' do
    begin
      filenamepart := filenamepart + filename[i] ;
      i := i + 1 ;
    end ;
  asmfilename := filenamepart + '.asm' ;
  assign (asmfile, asmfilename) ;
  rewrite(asmfile) ;
  listfilename := filenamepart + '.lst' ;
  assign(listfile, listfilename) ;
  rewrite(listfile) ;

END ; (* OpenFiles *)

PROCEDURE CloseFiles ;

  BEGIN

    close(objectfile) ;
    close(listfile) ;
    close(asmfile) ;

  END ; (* CloseFiles *)


BEGIN (* main PasDis *)

    writeln ;
    write( 'Disassembler for Pascal-M V1.4') ;



    listpages := 1 ;
    listlines := maxpage + 1 ;

    FillInstructions ;

    pc := ldaddr ;
    OpenFiles ;

    Loader ;
    if fstatus <> checkerror
      then
        BEGIN

          {
          disassemble(phase1) ;
          }
         disassemble { (phase2)} ;
         DumpProcedureBuffer ;
         writeln (filename, ' disassembled for load address ', hexstr(ldaddr,4))  ;
       end ;
     CloseFiles ;
END.


